<?php

namespace NoccyLabs\Plugin\Manager;


class PluginManager
{
    /** @var PluginManifest[] */
    private $plugins = [];
    /** @var mixed */
    private $context;

    /**
     * Manually add a plugin to the manager. The plugin will not be loaded automatically.
     * The argument can be a directory, a zip file or a PluginManifest.
     *
     * @param string|PluginManifest
     * @throws PluginManagerException
     */
    public function addPlugin($plugin)
    {
        // Check if the provided plugin is a string; if so look for a manifest.
        if (is_string($plugin)) {
            if (is_dir($plugin)) {
                if (!file_exists($plugin . "/plugin.json")) {
                    throw new PluginManagerException("Invalid or incorrect path");
                }
                $plugin = PluginManifest::createFromFile($plugin . "/plugin.json");
                $this->plugins[] = $plugin;
                return;
            } elseif (is_file($plugin)) {
                $this->plugins[] = $plugin;
                return;
            }
        }
        if ($plugin instanceof PluginManifest) {
            $this->plugins[] = $plugin;
            return;
        }
        throw new PluginManagerException("Expected PluginManifest or directory");
    }

    /**
     * Scan a directory looking for plugins
     *
     * @param string The path to scan
     */
    public function scanPlugins($path)
    {
        $it = new \DirectoryIterator($path);
        foreach ($it as $item) {
            try {
                /** @var \SplFileInfo $item */
                if ($item->isDir()) {
                    if (substr($item->getFilename(), 0, 1) == ".") {
                        continue;
                    }
                    if (!file_exists($item->getPathname() . "/plugin.json")) {
                        continue;
                    }
                    $this->scanPlugin($item->getPathname());
                } else {
                    if ($item->getExtension() == "zip") {
                        $this->scanPlugin($item->getPathname());
                    }
                }
            } catch (PluginManifestException $e) {
                @fprintf(STDERR, "%s: %s\n", $path, $e);
            }
        }
    }

    public function scanPlugin(string $path)
    {
        if (is_dir($path)) {
            $manifest = file_get_contents($path . "/plugin.json");
        } elseif (fnmatch("*.zip", $path)) {
            $manifest = @file_get_contents("zip://{$path}#plugin.json");
        }
        $manifest = json_decode($manifest);
        if (empty($manifest->name) || empty($manifest->version)) {
            throw new PluginManifestException("Invalid manifest in {$path}/plugin.json");
        }
        $plugin = PluginManifest::create($manifest->name, $manifest->version, (array)$manifest, $path);
        $this->addPlugin($plugin);
    }

    /**
     * Load a plugin (optionally overriding the context)
     *
     * @param string The plugin id to load, optionally @ver
     * @param null|mixed If not null, the context to provide the plugin when requested
     */
    public function loadPlugin($plugin, $context = null)
    {
        $manifest = $this->findPlugin($plugin);
        $requires = $manifest->getKey("requires");
        if (is_array($requires) && count($requires)) {
            foreach ($requires as $requirement) {
                $this->loadPlugin($requirement, $context);
            }
        }
    }

    /**
     * Find a plugin manifest
     *
     * @param string The plugin id to find, optionally @ver
     */
    public function findPlugin(string $plugin) : ? PluginManifest
    {
        if (strpos($plugin, '@') !== false) {
            list($plugin, $version) = explode("@", $plugin, 2);
        } else {
            $version = null;
        }

        foreach ($this->plugins as $manifest) {
            /** @var PluginManifest $manifest */
            if ($manifest->getName() != $plugin) {
                continue;
            }
            if ($version && ($manifest->compareVersion($version) !== 0)) {
                continue;
            }
            return $manifest;
        }
        return $null;
    }

    public function findInPlugins(string $path)
    {
        $results = [];
        foreach ($this->plugins as $manifest) {
            try {
                $pluginPath = $manifest->translatePath($path);
                if (file_exists($pluginPath))
                    $results[] = $manifest;
            } catch (PluginManifestException $e) {
                continue;
            }
        }
        return $results;
    }

    /**
     * Register stream wrappers, providing the plugin:// streams. After calling this function
     * you can access files relative to the plugin's root using plugin://id.of.plugin/path/file
     *
     */
    public function registerWrapper()
    {
        PluginStream::enable($this);
    }

    /**
     * Set the default context for this manager.
     *
     * @param null|mixed
     */
    public function setContext($context)
    {
        $this->context = $context;
    }

    /**
     *
     * @return null|mixed
     */
    public function getContext()
    {
        return $this->context;
    }

}
